'use strict';

function addheading(carousel, colspan, content)
{
	let thead = document.createElement("thead");
	let tr = document.createElement("tr");
	let th = document.createElement("th");
	th.innerHTML = content;
	th.setAttribute("colspan", colspan);
	tr.appendChild(th);
	thead.appendChild(tr);
	carousel.appendChild(thead);
};
function addtd(tr, className, content) {
	let td = document.createElement("td");
	td.appendChild(document.createTextNode(content));
	td.className = className;
	tr.appendChild(td);
};
function addth(tr, className, content) {
	let th = document.createElement("th");
	th.appendChild(document.createTextNode(content));
	th.className = className;
	tr.appendChild(th);
};

function subrank_partitions(games, parts, start_rank, tiebreakers) {
	let result = [];
	for (let i = 0; i < parts.length; ++i) {
		let part = rank(games, parts[i], start_rank, tiebreakers);
		for (let j = 0; j < part.length; ++j) {
			result.push(part[j]);
		}
		start_rank += part.length;
	}
	return result;
};

function partition(teams, compare)
{
	teams.sort(compare);

	let parts = [];
	let curr_part = [teams[0]];
	for (let i = 1; i < teams.length; ++i) {
		if (compare(teams[i], curr_part[0]) != 0) {
			parts.push(curr_part);
			curr_part = [];
		}
		curr_part.push(teams[i]);
	}
	if (curr_part.length != 0) {
		parts.push(curr_part);
	}
	return parts;
};

function explain_tiebreaker(parts, rule_name)
{
	let result = [];
	for (let i = 0; i < parts.length; ++i) {
		result.push(parts[i].map(function(x) { return x.shortname; }).join("/"));
	}
	return result.join(" > ") + " (" + rule_name + ")";
}

function make_teams_to_idx(teams)
{
	let teams_to_idx = [];
	for (let i = 0; i < teams.length; i++) {
		teams_to_idx[teams[i].name] = i;
	}
	return teams_to_idx;
}

function partition_by_beat(teams, fill_beatmatrix)
{
	// Head-to-head score by way of components. First construct the beat matrix.
	let n = teams.length;
	let beat = new Array(n);
	let teams_to_idx = make_teams_to_idx(teams);
	for (let i = 0; i < n; i++) {
		beat[i] = new Array(n);
		for (let j = 0; j < n; j++) {
			beat[i][j] = 0;
		}
	}
	fill_beatmatrix(beat, teams_to_idx);
	// Floyd-Warshall for transitive closure.
	for (let k = 0; k < n; ++k) {
		for (let i = 0; i < n; ++i) {
			for (let j = 0; j < n; ++j) {
				if (beat[i][k] && beat[k][j]) {
					beat[i][j] = 1;
				}
			}
		}
	}

	// See if we can find any team that is comparable to all others.
	for (let pivot_idx = 0; pivot_idx < n; pivot_idx++) {
		let incomparable = false;
		for (let i = 0; i < n; ++i) {
			if (i != pivot_idx && beat[pivot_idx][i] == 0 && beat[i][pivot_idx] == 0) {
				incomparable = true;
				break;
			}
		}
		if (!incomparable) {
			// Split the teams into three partitions:
			let better_than_pivot = [], equal = [], worse_than_pivot = [];
			for (let i = 0; i < n; ++i) {
				let we_beat = (beat[pivot_idx][i] == 1);
				let they_beat = (beat[i][pivot_idx] == 1);
				if ((i == pivot_idx) || (we_beat && they_beat)) {
					equal.push(teams[i]);
				} else if (we_beat && !they_beat) {
					worse_than_pivot.push(teams[i]);
				} else if (they_beat && !we_beat) {
					better_than_pivot.push(teams[i]);
				} else {
					console.log("this shouldn't happen");
				}
			} 
			let result = [];
			if (better_than_pivot.length > 0) {
				result = partition_by_beat(better_than_pivot, fill_beatmatrix);
			}
			result.push(equal);  // Obviously can't be partitioned further.
			if (worse_than_pivot.length > 0) {
				result = result.concat(partition_by_beat(worse_than_pivot, fill_beatmatrix));
			}
			return result;
		}
	}

	// No usable pivot was found, so the graph is inherently
	// disconnected, and we cannot partition it.
	return [teams];
}

// Takes in an array, gives every element a rank starting with 1, and returns.
function rank(games, teams, start_rank, tiebreakers) {
	if (teams.length <= 1) {
		// Only one team, so trivial.
		teams[0].rank = start_rank;
		return teams;
	}

	// Rule #0: Partition the teams by score.
	let score_parts = partition(teams, function(a, b) { return b.pts - a.pts });
	if (score_parts.length > 1) {
		return subrank_partitions(games, score_parts, start_rank, tiebreakers);
	}

	// Rule #1: Head-to-head wins.
	let beat_parts = partition_by_beat(teams, function(beat, teams_to_idx) {
		for (let i = 0; i < games.length; ++i) {
			let idx1 = teams_to_idx[games[i].name1];
			let idx2 = teams_to_idx[games[i].name2];
			if (idx1 !== undefined && idx2 !== undefined) {
				if (games[i].score1 > games[i].score2) {
					beat[idx1][idx2] = 1;
				}
				if (games[i].score1 < games[i].score2) {
					beat[idx2][idx1] = 1;
				}
			}
		}
	});
	if (beat_parts.length > 1) {
		tiebreakers.push(explain_tiebreaker(beat_parts, 'head-to-head'));
		return subrank_partitions(games, beat_parts, start_rank, tiebreakers);
	}

	// Rule #2: Number of games played (fewer is better).
	// Actually the rule says “fewest losses”, but fewer games is equivalent
	// as long as teams have the same amount of points and ties don't exist.
	let nplayed_parts = partition(teams, function(a, b) { return a.nplayed - b.nplayed });
	if (nplayed_parts.length > 1) {
		tiebreakers.push(explain_tiebreaker(nplayed_parts, 'fewer losses'));
		return subrank_partitions(games, nplayed_parts, start_rank, tiebreakers);
	}

	// Rule #3: Head-to-head goal difference. 
	let teams_to_idx = make_teams_to_idx(teams);
	for (let i = 0; i < teams.length; i++) {
		teams[i].h2h_gd = 0;
		teams[i].h2h_goals = 0;
	}
	for (let i = 0; i < games.length; ++i) {
		let idx1 = teams_to_idx[games[i].name1];
		let idx2 = teams_to_idx[games[i].name2];
		if (idx1 !== undefined && idx2 !== undefined &&
		    !isNaN(games[i].score1) && isNaN(games[i].score2)) {
			teams[idx1].h2h_gd += games[i].score1;
			teams[idx1].h2h_gd -= games[i].score2;
			teams[idx2].h2h_gd += games[i].score2;
			teams[idx2].h2h_gd -= games[i].score1;

			teams[idx1].h2h_goals += games[i].score1;
			teams[idx2].h2h_goals += games[i].score2;
		}
	}
	let h2h_gd_parts = partition(teams, function(a, b) { return b.h2h_gd - a.h2h_gd });
	if (h2h_gd_parts.length > 1) {
		tiebreakers.push(explain_tiebreaker(h2h_gd_parts, 'head-to-head goal difference'));
		return subrank_partitions(games, h2h_gd_parts, start_rank, tiebreakers);
	}

	// Rule #4: Goal difference against common opponents.
	var results = {};
	for (let i = 0; i < games.length; ++i) {
		if (results[games[i].name1] === undefined) {
			results[games[i].name1] = {};
		}
		if (results[games[i].name2] === undefined) {
			results[games[i].name2] = {};
		}
		results[games[i].name1][games[i].name2] = [ games[i].score1, games[i].score2 ];
		results[games[i].name2][games[i].name1] = [ games[i].score2, games[i].score1 ];
	}
	let gd_parts = partition_by_beat(teams, function(beat, teams_to_idx) {
		for (const team_i of Object.keys(teams_to_idx)) {
			let i = teams_to_idx[team_i];
			for (const team_j of Object.keys(teams_to_idx)) {
				let j = teams_to_idx[team_j];
				let results_i = results[team_i], results_j = results[team_j];
				let gd_i = 0, gd_j = 0;

				// See if the two teams have both played a third team k.
				for (let k in results_i) {
					if (!results_i.hasOwnProperty(k)) continue;
					if (results_j[k] !== undefined) {
						gd_i += results_i[k][0] - results_i[k][1];
						gd_j += results_j[k][0] - results_j[k][1];
					}
				}

				if (gd_i > gd_j) {
					beat[i][j] = 1;
				} else if (gd_i < gd_j) {
					beat[j][i] = 1;
				}
			}
		}
	});
	if (gd_parts.length > 1) {
		tiebreakers.push(explain_tiebreaker(gd_parts, 'goal difference versus common opponents'));
		return subrank_partitions(games, gd_parts, start_rank, tiebreakers);
	}

	// Rule #5: Head-to-head scored goals.
	let h2h_goals_parts = partition(teams, function(a, b) { return b.h2h_goals - a.h2h_goals });
	if (h2h_goals_parts.length > 1) {
		tiebreakers.push(explain_tiebreaker(h2h_goals_parts, 'head-to-head scored goals'));
		return subrank_partitions(games, h2h_goals_parts, start_rank, tiebreakers);
	}

	// Rule #6: Goals scored against common opponents.
	let goals_parts = partition_by_beat(teams, function(beat, teams_to_idx) {
		for (const team_i of Object.keys(teams_to_idx)) {
			let i = teams_to_idx[team_i];
			for (const team_j of Object.keys(teams_to_idx)) {
				let j = teams_to_idx[team_j];
				let results_i = results[team_i], results_j = results[team_j];
				let goals_i = 0, goals_j = 0;

				// See if the two teams have both played a third team k.
				for (let k in results_i) {
					if (!results_i.hasOwnProperty(k)) continue;
					if (results_j[k] !== undefined) {
						goals_i += results_i[k][0];
						goals_j += results_j[k][0];
					}
				}

				if (goals_i > goals_j) {
					beat[i][j] = 1;
				} else if (goals_i < goals_j) {
					beat[j][i] = 1;
				}
			}
		}
	});
	if (goals_parts.length > 1) {
		tiebreakers.push(explain_tiebreaker(goals_parts, 'goals scored against common opponents'));
		return subrank_partitions(games, goals_parts, start_rank, tiebreakers);
	}

	// OK, it's a tie. Give them all the same rank.
	let result = [];
	for (let i = 0; i < teams.length; ++i) {
		result.push(teams[i]);
		result[i].rank = start_rank;
	}
	return result; 
}; 

function parse_teams_from_spreadsheet(response) {
	let teams = [];
	for (let i = 2; response.values[i].length >= 1; ++i) {
		teams.push({
			"name": response.values[i][0],
			"mediumname": response.values[i][1],
			"shortname": response.values[i][2],
			"tags": response.values[i][3],
			"nplayed": 0,
			"gd": 0,
			"pts": 0,
			"goals": 0
		});
	}
	return teams;
};

function parse_games_from_spreadsheet(response, group_name, include_unplayed) {
	let games = [];
	let i;
	for (i = 0; i < response.values.length; ++i) {
		if (response.values[i][0] === 'Results') {
			i += 2;
			break;
		}
	}

	for ( ; response.values[i] !== undefined && response.values[i].length >= 1; ++i) {
		if ((response.values[i][2] && response.values[i][3]) || include_unplayed) {
			let real_group_name = response.values[i][9];
			if (real_group_name === undefined) {
				real_group_name = group_name;
			}
			games.push({
				"name1": response.values[i][0],
				"name2": response.values[i][1],
				"score1": parseInt(response.values[i][2]),
				"score2": parseInt(response.values[i][3]),
				"streamday": response.values[i][7],
				"streamtime": response.values[i][8],
				"group_name": real_group_name
			});
		}
	}
	return games;
};

function display_group(response, group_name)
{
	let teams = parse_teams_from_spreadsheet(response);
	let games = parse_games_from_spreadsheet(response, group_name, false);
	display_group_parsed(teams, games, group_name);
};

function display_group_parsed(teams, games, group_name)
{
	document.getElementById('entire-bug').style.display = 'none';

	let teams_to_idx = make_teams_to_idx(teams);
	for (let i = 0; i < games.length; ++i) {
		let idx1 = teams_to_idx[games[i].name1];
		let idx2 = teams_to_idx[games[i].name2];
		if (games[i].score1 === undefined || games[i].score2 === undefined ||
		    isNaN(games[i].score1) || isNaN(games[i].score2) ||
		    idx1 === undefined || idx2 === undefined ||
		    games[i].score1 == games[i].score2) {
			continue;
		}
		++teams[idx1].nplayed;
		++teams[idx2].nplayed;
		teams[idx1].goals += games[i].score1;
		teams[idx2].goals += games[i].score2;
		teams[idx1].gd += games[i].score1;
		teams[idx2].gd += games[i].score2;
		teams[idx1].gd -= games[i].score2;
		teams[idx2].gd -= games[i].score1;
		if (games[i].score1 > games[i].score2) {
			teams[idx1].pts += 2;
		} else {
			teams[idx2].pts += 2;
		}
	}

	let tiebreakers = [];
	teams = rank(games, teams, 1, tiebreakers);

	let carousel = document.getElementById('carousel');
	clear_carousel(carousel);

	addheading(carousel, 5, "Current standings, Trøndisk 2017<br />" + group_name);
	let tr = document.createElement("tr");
	tr.className = "subfooter";
	addth(tr, "rank", "");
	addth(tr, "team", "");
	addth(tr, "nplayed", "P");
	addth(tr, "gd", "GD");
	addth(tr, "pts", "Pts");
	carousel.appendChild(tr);

	let row_num = 2;
	for (let i = 0; i < teams.length; ++i) {
		let tr = document.createElement("tr");

		addth(tr, "rank", teams[i].rank);
		addtd(tr, "team", teams[i].name);
		addtd(tr, "nplayed", teams[i].nplayed);
		addtd(tr, "gd", teams[i].gd.toString().replace(/-/, '−'));
		addtd(tr, "pts", teams[i].pts);

		carousel.appendChild(tr);
	}

	if (tiebreakers.length > 0) {
		let tie_tr = document.createElement("tr");
		tie_tr.className = "footer";
		let td = document.createElement("td");
		td.appendChild(document.createTextNode("Tiebreaks applied: " + tiebreakers.join(', ')));
		td.setAttribute("colspan", "5");
		tie_tr.appendChild(td);
		carousel.appendChild(tie_tr);
	}

	let footer_tr = document.createElement("tr");
	footer_tr.className = "footer";
	let td = document.createElement("td");
	td.appendChild(document.createTextNode("www.trondheimfrisbeeklubb.no | #trøndisk"));
	td.setAttribute("colspan", "5");
	footer_tr.appendChild(td);
	carousel.appendChild(footer_tr);

	fade_in_rows(carousel);

	carousel.style.display = 'table';
};

function fade_in_rows(table)
{
	let trs = table.getElementsByTagName("tr");
	for (let i = 1; i < trs.length; ++i) {  // The header already has its own fade-in.
		if (trs[i].className === "footer") {
			trs[i].style = "-webkit-animation: fade-in 1.0s ease; -webkit-animation-delay: " + (0.25 * i) + "s; -webkit-animation-fill-mode: both;";
		} else {
			trs[i].style = "-webkit-animation: fade-in 2.0s ease; -webkit-animation-delay: " + (0.25 * i) + "s; -webkit-animation-fill-mode: both;";
		}
	}
};

function fade_out_rows(table)
{
	let trs = table.getElementsByTagName("tr");
	for (let i = 0; i < trs.length; ++i) {
		if (trs[i].className === "footer") {
			trs[i].style = "-webkit-animation: fade-out 1.0s ease; -webkit-animation-delay: " + (0.125 * i) + "s; -webkit-animation-fill-mode: both;";
		} else {
			trs[i].style = "-webkit-animation: fade-out 1.0s ease; -webkit-animation-delay: " + (0.125 * i) + "s; -webkit-animation-fill-mode: both;";
		}
	}
};

function clear_carousel(table)
{
	while (table.childNodes.length > 0) {
		table.removeChild(table.firstChild);
	}
};

// Stream schedule
let max_list_len = 8;

function display_stream_schedule(response, group_name) {
	let teams = parse_teams_from_spreadsheet(response);
	let games = parse_games_from_spreadsheet(response, group_name, true);
	display_stream_schedule_parsed(teams, games, 0);
};

function sort_game_list(games) {
	games = games.filter(function(game) { return game.streamtime !== undefined && game.streamtime.match(/[0-9]+:[0-9]+/) != null; });
	games.sort(function(a, b) {
		if (a.streamday !== b.streamday) {
			return a.streamday - b.streamday;
		}

		let m1 = a.streamtime.match(/([0-9]+):([0-9]+)/);
		let m2 = b.streamtime.match(/([0-9]+):([0-9]+)/);
		return (m1[1] * 60 + m1[2]) - (m2[1] * 60 + m2[2]);
	});
	return games;
}

function find_game_start_idx(games) {
	// Pick out a reasonable place to start the list. We'll show the last
	// completed match and start from there.
	let start_idx = games.length - 1;
	for (let i = 0; i < games.length; ++i) {
		if (isNaN(games[i].score1) || isNaN(games[i].score2) &&
		    games[i].score1 === games[i].score2) {
			start_idx = i;
			break;
		}
	}
	if (start_idx > 0) start_idx--;
	if (games.length >= max_list_len) {
		start_idx = Math.min(start_idx, games.length - max_list_len);
	}
	return start_idx;
}

function find_num_pages(games) {
	games = sort_game_list(games);
	let start_idx = find_game_start_idx(games);
	return Math.ceil((games.length - start_idx) / max_list_len);
}

function display_stream_schedule_parsed(teams, games, page) {
	document.getElementById('entire-bug').style.display = 'none';

	games = sort_game_list(games);
	let start_idx = find_game_start_idx(games);

	start_idx += page * max_list_len;
	if (start_idx >= games.length) {
		// Error.
		return;
	}

	let days = ["Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday", "Sunday"];
	let shortdays = ["Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun"];
	let today = days[(new Date).getDay()];

	let covered_days = [];
	let row_num = 0;
	for (let i = start_idx; i < games.length && row_num++ < max_list_len; ++i) {
		if (i == start_idx || games[i].streamday != games[i - 1].streamday) {
			covered_days.push(days[games[i].streamday]);
		}
	}
	
	let carousel = document.getElementById('carousel');
	clear_carousel(carousel);
	addheading(carousel, 3, "Stream schedule, Trøndisk 2017<br />" + covered_days.join('/') + " (all times CET)");

	let teams_to_idx = make_teams_to_idx(teams);
	row_num = 0;
	for (let i = start_idx; i < games.length && row_num < max_list_len; ++i) {
		let tr = document.createElement("tr");

		let name1 = teams[teams_to_idx[games[i].name1]].mediumname;
		let name2 = teams[teams_to_idx[games[i].name2]].mediumname;

		addtd(tr, "matchup", name1 + "–" + name2);
		addtd(tr, "group", games[i].group_name);

		if (!isNaN(games[i].score1) && !isNaN(games[i].score2) &&
		    games[i].score1 !== games[i].score2) {
			addtd(tr, "streamtime", games[i].score1 + "–" + games[i].score2);
		} else {
			let streamtime = games[i].streamtime;
			let streamday = days[games[i].streamday];
			if (streamday !== today) {
				streamtime = shortdays[games[i].streamday] + " " + streamtime;
			}
			addth(tr, "streamtime", streamtime);
		}

		row_num++;
		carousel.appendChild(tr);
	}

	fade_in_rows(carousel);

	carousel.style.display = 'table';
};

function get_group(group_name, cb)
{
	let req = new XMLHttpRequest();
	req.onload = function(e) {
		cb(JSON.parse(req.responseText), group_name);
	};
	req.open('GET', 'https://sheets.googleapis.com/v4/spreadsheets/122tIwrXTi5ug0Vv6Np5w3pVwEWE2KkjWxtzQQfGtOZA/values/\'' + group_name + '\'!A1:J50?key=AIzaSyAuP9yQn8g0bSay6r_RpGtpFeIbwprH1TU');
	req.send();
};

function showgroup(group_name)
{
	get_group(group_name, display_group);
};

function showgroup_from_state()
{
	showgroup(state['group_name']);
};

let carousel_timeout = null;

function hidetable()
{
	fade_out_rows(document.getElementById('carousel'));
};

function showschedule(page)
{
	let teams = [];
	let games = [];
	let num_left = 3;

	let cb = function(response, group_name) {
		teams = teams.concat(parse_teams_from_spreadsheet(response));
		games = games.concat(parse_games_from_spreadsheet(response, group_name, true));
		if (--num_left == 0) {
			display_stream_schedule_parsed(teams, games, 0);
		}
	};

	get_group('Group A', cb);
	get_group('Group B', cb);
	get_group('Playoffs', cb);
};

function do_series(series)
{
	do_series_internal(series, 0);
};

function do_series_internal(series, idx)
{
	(series[idx][1])();
	if (idx + 1 < series.length) {
		carousel_timeout = setTimeout(function() { do_series_internal(series, idx + 1); }, series[idx][0]);
	}
};

function showcarousel()
{
	let teams_per_group = [];
	let games_per_group = [];
	let combined_teams = [];
	let combined_games = [];
	let num_left = 3;

	let cb = function(response, group_name) {
		let teams = parse_teams_from_spreadsheet(response);
		let games = parse_games_from_spreadsheet(response, group_name, true);
		teams_per_group[group_name] = teams;
		games_per_group[group_name] = games;

		combined_teams = combined_teams.concat(teams);
		combined_games = combined_games.concat(games);
		if (--num_left == 0) {
			let series = [
				[ 13000, function() { display_group_parsed(teams_per_group['Group A'], games_per_group['Group A'], 'Group A'); } ],
				[ 2000, function() { hidetable(); } ],
				[ 13000, function() { display_group_parsed(teams_per_group['Group B'], games_per_group['Group B'], 'Group B'); } ],
				[ 2000, function() { hidetable(); } ]
			];
			let num_pages = find_num_pages(combined_games);
			for (let page = 0; page < num_pages; ++page) {
				series.push([ 13000, function() { display_stream_schedule_parsed(combined_teams, combined_games, page); } ]);
				series.push([ 2000, function() { hidetable(); } ]);
			}

			do_series(series);
		}
	};

	get_group('Group A', cb);
	get_group('Group B', cb);
	get_group('Playoffs', cb);
};

function stopcarousel()
{
	if (carousel_timeout !== null) {
		hidetable();
		clearTimeout(carousel_timeout);
		carousel_timeout = null;
	}
};

function hidescorebug()
{
	document.getElementById('entire-bug').style.display = 'none';
}

function showscorebug()
{
	document.getElementById('entire-bug').style.display = null;
};

